@@ -12,6 +12,21 @@
 #include <string.h>
 #include <unistd.h>
 
+/******************** Constants ********************/
+/* HTTP timeouts, by operation, in milliseconds
+ */
+#define DEVICE_HTTP_TIMEOUT_DEVCAPS     5000
+#define DEVICE_HTTP_TIMEOUT_SCAN        30000
+#define DEVICE_HTTP_TIMEOUT_LOAD        -1
+#define DEVICE_HTTP_TIMEOUT_CHECK       5000
+#define DEVICE_HTTP_TIMEOUT_CLEANUP     30000
+#define DEVICE_HTTP_TIMEOUT_CANCEL      30000
+
+/* HTTP timeout for operation that was pending
+ * in a moment of cancel (if any)
+ */
+#define DEVICE_HTTP_TIMEOUT_CANCELED_OP 10000
+
 /******************** Device management ********************/
 /* Device flags
  */
@@ -212,7 +227,7 @@ device_new (zeroconf_devinfo *devinfo)
 /* Destroy a device
  */
 static void
-device_free (device *dev)
+device_free (device *dev, const char *log_msg)
 {
     int i;
 
@@ -255,6 +270,10 @@ device_free (device *dev)
     pollable_free(dev->read_pollable);
 
     log_debug(dev->log, "device destroyed");
+    if (log_msg != NULL) {
+        log_debug(dev->log, "%s", log_msg);
+    }
+
     log_ctx_free(dev->log);
     zeroconf_devinfo_free(dev->devinfo);
     mem_free(dev);
@@ -277,7 +296,6 @@ device_io_start (device *dev)
 {
     dev->stm_cancel_event = eloop_event_new(device_stm_cancel_event_callback, dev);
     if (dev->stm_cancel_event == NULL) {
-        device_close(dev);
         return SANE_STATUS_NO_MEM;
     }
 
@@ -310,7 +328,7 @@ static void
 device_table_purge (void)
 {
     while (mem_len(device_table) > 0) {
-        device_free(device_table[0]);
+        device_free(device_table[0], NULL);
     }
 }
 
@@ -356,6 +374,7 @@ device_proto_devcaps_submit (device *dev, void (*callback) (void*, http_query*))
     http_query *q;
 
     q = dev->proto_ctx.proto->devcaps_query(&dev->proto_ctx);
+    http_query_timeout(q, DEVICE_HTTP_TIMEOUT_DEVCAPS);
     http_query_submit(q, callback);
     dev->proto_ctx.query = q;
 }
@@ -386,6 +405,18 @@ device_proto_op_name (device *dev, PROTO_OP op)
     return NULL;
 }
 
+/* http_query_onrxhdr() callback
+ */
+static void
+device_proto_op_onrxhdr (void *p, http_query *q)
+{
+    device *dev = p;
+
+    if (dev->proto_op_current == PROTO_OP_LOAD && !dev->stm_cancel_sent) {
+        http_query_timeout(q, -1);
+    }
+}
+
 /* Submit operation request
  */
 static void
@@ -393,15 +424,32 @@ device_proto_op_submit (device *dev, PROTO_OP op,
         void (*callback) (void*, http_query*))
 {
     http_query *(*func) (const proto_ctx *ctx) = NULL;
+    int        timeout = -1;
     http_query *q;
 
     switch (op) {
     case PROTO_OP_NONE:    log_internal_error(dev->log); break;
-    case PROTO_OP_SCAN:    func = dev->proto_ctx.proto->scan_query; break;
-    case PROTO_OP_LOAD:    func = dev->proto_ctx.proto->load_query; break;
-    case PROTO_OP_CHECK:   func = dev->proto_ctx.proto->status_query; break;
-    case PROTO_OP_CLEANUP: func = dev->proto_ctx.proto->cleanup_query; break;
     case PROTO_OP_FINISH:  log_internal_error(dev->log); break;
+
+    case PROTO_OP_SCAN:
+        func = dev->proto_ctx.proto->scan_query;
+        timeout = DEVICE_HTTP_TIMEOUT_SCAN;
+        break;
+
+    case PROTO_OP_LOAD:
+        func = dev->proto_ctx.proto->load_query;
+        timeout = DEVICE_HTTP_TIMEOUT_LOAD;
+        break;
+
+    case PROTO_OP_CHECK:
+        func = dev->proto_ctx.proto->status_query;
+        timeout = DEVICE_HTTP_TIMEOUT_CHECK;
+        break;
+
+    case PROTO_OP_CLEANUP:
+        func = dev->proto_ctx.proto->cleanup_query;
+        timeout = DEVICE_HTTP_TIMEOUT_CLEANUP;
+        break;
     }
 
     log_assert(dev->log, func != NULL);
@@ -409,7 +457,13 @@ device_proto_op_submit (device *dev, PROTO_OP op,
     log_debug(dev->log, "%s: submitting: attempt=%d",
         device_proto_op_name(dev, op), dev->proto_ctx.failed_attempt);
     dev->proto_op_current = op;
+
     q = func(&dev->proto_ctx);
+    http_query_timeout(q, timeout);
+    if (op == PROTO_OP_LOAD) {
+        http_query_onrxhdr(q, device_proto_op_onrxhdr);
+    }
+
     http_query_submit(q, callback);
     dev->proto_ctx.query = q;
 }
@@ -724,9 +778,12 @@ device_stm_cancel_perform (device *dev, SANE_Status status)
             ctx->images_received > 0) {
             /* If we are not expecting more images, skip cancel
              * and simple wait until job is done
+             *
+             * Otherwise Xerox VersaLink B405 remains busy for
+             * a quite long time without any need
              */
             log_debug(dev->log, "cancel skipped as job is almost done");
-            device_stm_state_set(dev, DEVICE_STM_CANCEL_REQ_DONE);
+            return false;
         } else {
             /* Otherwise, perform a normal cancel operation
              */
@@ -736,6 +793,12 @@ device_stm_cancel_perform (device *dev, SANE_Status status)
             dev->stm_cancel_query = ctx->proto->cancel_query(ctx);
 
             http_query_onerror(dev->stm_cancel_query, NULL);
+            http_query_timeout(dev->stm_cancel_query,
+                    DEVICE_HTTP_TIMEOUT_CANCEL);
+
+            http_client_timeout(dev->proto_ctx.http,
+                    DEVICE_HTTP_TIMEOUT_CANCELED_OP);
+
             http_query_submit(dev->stm_cancel_query, device_stm_cancel_callback);
 
             dev->stm_cancel_sent = true;
@@ -753,22 +816,28 @@ device_stm_cancel_event_callback (void *data)
 {
     device *dev = data;
 
-    log_debug(dev->log, "cancel requested");
+    log_debug(dev->log, "cancel processing started");
     if (!device_stm_cancel_perform(dev, SANE_STATUS_CANCELLED)) {
         device_stm_state_set(dev, DEVICE_STM_CANCEL_DELAYED);
     }
 }
 
-/* Request cancel
+/* Request cancel.
+ *
+ * Note, reason must be NULL, if cancel requested from the signal handler
  */
 static void
-device_stm_cancel_req (device *dev)
+device_stm_cancel_req (device *dev, const char *reason)
 {
     DEVICE_STM_STATE expected = DEVICE_STM_SCANNING;
     bool ok = __atomic_compare_exchange_n(&dev->stm_state, &expected,
         DEVICE_STM_CANCEL_REQ, true, __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST);
 
     if (ok) {
+        if (reason != NULL) {
+            log_debug(dev->log, "cancel requested: %s", reason);
+        }
+
         eloop_event_trigger(dev->stm_cancel_event);
     }
 }
@@ -1044,9 +1113,9 @@ device_stm_wait_while_working (device *dev)
 /* Cancel scanning and wait until device leaves the working state
  */
 static void
-device_stm_cancel_wait (device *dev)
+device_stm_cancel_wait (device *dev, const char *reason)
 {
-    device_stm_cancel_req(dev);
+    device_stm_cancel_req(dev, reason);
     device_stm_wait_while_working(dev);
 }
 
@@ -1130,8 +1199,8 @@ device_open (const char *ident, SANE_Status *status)
     dev = device_new(devinfo);
     *status = device_io_start(dev);
     if (*status != SANE_STATUS_GOOD) {
-        device_free(dev);
-        dev = NULL;
+        device_free(dev, NULL);
+        return NULL;
     }
 
     /* Wait until device is initialized */
@@ -1140,27 +1209,28 @@ device_open (const char *ident, SANE_Status *status)
     }
 
     if (device_stm_state_get(dev) == DEVICE_STM_PROBING_FAILED) {
-        device_free(dev);
-        dev = NULL;
+        device_free(dev, NULL);
         *status = SANE_STATUS_IO_ERROR;
+        return NULL;
     }
 
     return dev;
 }
 
 /* Close the device
+ * If log_msg is not NULL, it is written to the device log as late as possible
  */
 void
-device_close (device *dev)
+device_close (device *dev, const char *log_msg)
 {
     /* Cancel job in progress, if any */
     if (device_stm_state_working(dev)) {
-        device_stm_cancel_wait(dev);
+        device_stm_cancel_wait(dev, "device close");
     }
 
     /* Close the device */
     device_stm_state_set(dev, DEVICE_STM_CLOSED);
-    device_free(dev);
+    device_free(dev, log_msg);
 }
 
 /* Get option descriptor
@@ -1307,6 +1377,9 @@ device_start_new_job (device *dev)
         dev->flags |= DEVICE_READING;
     } else {
         dev->flags &= ~DEVICE_SCANNING;
+        if (device_stm_state_get(dev) == DEVICE_STM_DONE) {
+            device_stm_state_set(dev, DEVICE_STM_IDLE);
+        }
     }
 
     return status;
@@ -1343,6 +1416,7 @@ device_start (device *dev)
      */
     while (device_stm_state_working(dev)
         && http_data_queue_len(dev->read_queue) == 0) {
+        log_debug(dev->log, "device_start: waiting for background scan job");
         eloop_cond_wait(&dev->stm_cond);
     }
 
@@ -1351,6 +1425,7 @@ device_start (device *dev)
      */
     if (http_data_queue_len(dev->read_queue) > 0) {
         dev->flags |= DEVICE_READING;
+        pollable_signal(dev->read_pollable);
         return SANE_STATUS_GOOD;
     }
 
@@ -1377,7 +1452,17 @@ device_start (device *dev)
 void
 device_cancel (device *dev)
 {
-    device_stm_cancel_req(dev);
+    /* Note, xsane calls sane_cancel() after each successful
+     * scan "just in case", which kills scan job running in
+     * background. So ignore cancel request, if from the API
+     * point of view we are not "scanning" (i.e., not between
+     * sane_start() and sane_read() completion)
+     */
+    if ((dev->flags & DEVICE_SCANNING) == 0) {
+        return;
+    }
+
+    device_stm_cancel_req(dev, NULL);
 }
 
 /* Set I/O mode
@@ -1717,7 +1802,7 @@ device_read (device *dev, SANE_Byte *data, SANE_Int max_len, SANE_Int *len_out)
 
     if (status == SANE_STATUS_IO_ERROR) {
         device_job_set_status(dev, SANE_STATUS_IO_ERROR);
-        device_cancel(dev);
+        device_stm_cancel_req(dev, "I/O error");
     }
 
     /* Cleanup and exit */
