@@ -96,6 +96,11 @@ trace_open (const char *device_name)
 
     strcpy(path, conf.dbg_trace);
     len = strlen(path);
+    if (len != 0 && path[len - 1] != '/') {
+        path[len ++] = '/';
+        path[len] = '\0';
+    }
+
     strcat(path, program);
     strcat(path, "-");
     strcat(path, device_name);
@@ -204,7 +209,7 @@ trace_dump_data (trace *t, http_data *data)
     strcpy(hdr.uid, "0");
     strcpy(hdr.gid, "0");
     sprintf(hdr.size, "%lo", (unsigned long) data->size);
-    sprintf(hdr.mtime, "%lo", time(NULL));
+    sprintf(hdr.mtime, "%llo", (long long) time(NULL));
     hdr.typeflag[0] = '0';
     strcpy(hdr.magic, "ustar");
     memcpy(hdr.version, "00", 2);
